#include <stdio.h>
#include <stdlib.h>
#include <time.h>
// #include <sys/time.h>
#include <math.h>
#ifndef M_PI
#define M_PI 3.14159265
#endif

#define TILE_SIZE 32
__global__ 
void matrixMultKernel_global(float* Ad, float* Bd, float* Cd, int n);
__global__ 
void matrixMultKernel_tiled(float* Ad, float* Bd, float* Cd, int n);
__global__ 
void matrixMultKernel_coalesced(float* Ad, float* Bd, float* Cd, int n);
__global__ 
void matrixMultKernel_overlap(float* Ad, float* Bd, float* Cd, int n);
// define macro OUTPUT to print input & output matrix
// #define OUTPUT

void checkCUDAError(const char *msg);

void initMatrix(float *A, int n);
void dstMatrix(float *A, int n);
void matrixMult(float *A, float *B, float *C, int n);
void CUDA_matrixMult(float *A, float *B, float *C, int n, int repeats);
float* CUDA_initMatrix(float *A, int n);
void CUDA_freeMatrix(float *Ad);
void printMatrix(char* name, float *A, int n);

int main(int argc, char *argv[])
{
   float *A,*B,*C; /* arrays for matrices */
   int n, m; /* n=matrix size */
   
   clock_t clock1, clock2;
   float cpu_time = 0;

   if (argc < 3) {
      printf("Error: please specify matrix size and number of multiplications: \n");
      printf("Usage: %s 10 5 \n", argv[0]);
      exit(1);      
   };
   
   /* read block size and number of blocks */
   n = atoi(argv[1]);
   m = atoi(argv[2]);

   if (n % TILE_SIZE != 0) {
      printf("Error: matrix size has to be a multiple of tile size %d \n",
              TILE_SIZE);
      exit(1);      
   };

   printf("Matrix mult. of size %d (%d repeats): \n", n,m);

   /*
    * allocate and initialise matrices in host memory:
    */
   A = (float *) malloc(n*n*sizeof(float));
   dstMatrix(A,n);
   B = (float *) malloc(n*n*sizeof(float));
   dstMatrix(B,n);
   C = (float *) malloc(n*n*sizeof(float));
   initMatrix(C,n);
   
#ifdef OUTPUT
   printMatrix("A",A,n); printMatrix("B",B,n); 
   /* printMatrix("C",C,n); */ 
#endif

   printf("Start matrix multiplication:\n");
   clock1 = clock();
   cpu_time = 0.0;

   /* init Matrices */
   float* Ad = CUDA_initMatrix(A,n);
   float* Bd = CUDA_initMatrix(B,n);
   float* Cd = CUDA_initMatrix(C,n);

   /* perform matrix multiplication (m repeats) */
   CUDA_matrixMult(Ad,Bd,Cd,n,m);

   clock2 = clock();
   cpu_time += (clock2-clock1)/(float)CLOCKS_PER_SEC;

#ifdef OUTPUT
   printMatrix("Result Matrix: C",C,n);
#endif

   printf("Execution time: %f (CPU sec) \n",cpu_time);
   printf("Performance: %E Flop/s \n",m*pow(n,3.0)/cpu_time);

   int size = n*n*sizeof(float);
   cudaMemcpy(C,Cd, size, cudaMemcpyDeviceToHost);
      checkCUDAError("memory of C not transferred back");

   cudaFree(Ad);
   cudaFree(Bd);
   cudaFree(Cd);

   free(A);
   free(B);
   free(C);
   
   return(0);
}

/* set Matrix values to zero */
void initMatrix(float *A, int n)
{
   int i,k;

   for (i=0; i<n; i++)
     for (k=0; k<n; k++)
	A[i*n+k] = 0;
}

/* initialise Matrix: discrete Sine Transform */
void dstMatrix(float *A, int n)
{
   int i,k;

   for (i=0; i<n; i++)
     for (k=0; k<n; k++)
	A[i*n+k] = sin( (i*k*M_PI)/n );
}

/* 
 * matrix multiplication C += A*B 
 *  -> standard C implementation 
 */
void matrixMult(float *A, float *B, float *C, int n)
{
   int i,j,k;

   for (i=0; i<n; i++)
     for (k=0; k<n; k++)
     {
	for (j=0; j<n; j++)
           C[i*n+k] += A[i*n+j]*B[j*n+k];
     }
}



/* 
 * matrix initialization 
 */
__host__ 
float* CUDA_initMatrix(float *M, int n)
{
   int size = n*n*sizeof(float);
   float* Md;
   
   cudaMalloc((void**)&Md, size); 
      checkCUDAError("allocate memory for Matrix");

   cudaMemcpy(Md,M, size, cudaMemcpyHostToDevice);
      checkCUDAError("memory of A not transferred");

   return Md;
}

/* 
 * matrix multiplication C += A*B 
 *  -> CUDA implementation: kernel invocation
 *     (implementation adopted from Kirk&Hwu: 
 *      "Programming Massively Parallel Processors, chapter 3)
 */
__host__ void CUDA_matrixMult(float *Ad, float *Bd, float *Cd, int n, int repeats)
{
   int size = n*n*sizeof(float);
   
   dim3 dimBlock(TILE_SIZE,TILE_SIZE);
   dim3 dimGrid(n/TILE_SIZE,n/TILE_SIZE);
   
   for(int i=0; i<repeats; i++) 
 //     matrixMultKernel_global<<<dimGrid,dimBlock>>>(Ad,Bd,Cd,n);
 //     matrixMultKernel_tiled<<<dimGrid,dimBlock>>>(Ad,Bd,Cd,n);
      matrixMultKernel_coalesced<<<dimGrid,dimBlock>>>(Ad,Bd,Cd,n);
 //     matrixMultKernel_overlap<<<dimGrid,dimBlock>>>(Ad,Bd,Cd,n);
   
   checkCUDAError("matrix multiplication kernel failed");
}

/* print Matrix */
void printMatrix(char* name, float *A, int n)
{
   int i,k;

   printf("Matrix %s (size %d)\n",name,n);
   for (i=0; i<n; i++) {
     for (k=0; k<n; k++)
	printf("%f ", A[i*n+k]);
     printf("\n");
   };
}

/*
 * helper function to check for errors in CUDA calls
 * source: NVIDIA
 */
void checkCUDAError(const char *msg)
{
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err)
    {
        fprintf(stderr, "\nCuda error (%s): %s.\n", msg, cudaGetErrorString( err) );
        exit(-1);
    }
}

/* 
 * matrix multiplication C += A*B 
 *  -> CUDA kernel
 *     (implementation adopted from Kirk&Hwu: 
 *      "Programming Massively Parallel Processors, chapter 4)
 *  -> Features: none (basic tiled version, using only global memory)
 */
__global__ void matrixMultKernel_global(float* Ad, float* Bd, float* Cd, int n)
{
   int i = blockIdx.x * TILE_SIZE + threadIdx.x;
   int k = blockIdx.y * TILE_SIZE + threadIdx.y;
   
   float Celem = 0;
   
   for(int j=0; j<n; j++) {
      float Aelem = Ad[i*n+j];
      float Belem = Bd[j*n+k];
      Celem += Aelem*Belem;
   };
   
   Cd[i*n+k] += Celem;
}


/* 
 * matrix multiplication C += A*B 
 *  -> CUDA kernel
 *     (implementation adopted from Kirk&Hwu: 
 *      "Programming Massively Parallel Processors, chapter 5)
 *  -> Features:
 *     - tiled matrix multiplication with use of shared memory
 */
__global__ void matrixMultKernel_tiled(float* Ad, float* Bd, float* Cd, int n)
{
   __shared__ float Ads[TILE_SIZE][TILE_SIZE];
   __shared__ float Bds[TILE_SIZE][TILE_SIZE];

   int tx = threadIdx.x;
   int ty = threadIdx.y;
   
   int i = blockIdx.x * TILE_SIZE + tx;
   int k = blockIdx.y * TILE_SIZE + ty;
   
   float Celem = 0;
   
   for(int m=0; m < n/TILE_SIZE; m++) {
      Ads[tx][ty] = Ad[ i*n + m*TILE_SIZE+ty];
      Bds[tx][ty] = Bd[ (m*TILE_SIZE+tx)*n + k];
      __syncthreads();
      
      for(int j=0; j<TILE_SIZE; j++)
	 Celem += Ads[tx][j]*Bds[j][ty];
   
      __syncthreads();
   };
   Cd[i*n+k] += Celem;
}


/* 
 * matrix multiplication C += A*B 
 *  -> CUDA kernel
 *     (implementation adopted from Kirk&Hwu: 
 *      "Programming Massively Parallel Processors, chapter 5)
 *  -> Features:
 *     - tiled matrix multiplication with use of shared memory
 *     - coalesced memory access
 */
__global__ void matrixMultKernel_coalesced(float* Ad, float* Bd, float* Cd, int n)
{
   __shared__ float Ads[TILE_SIZE][TILE_SIZE];
   __shared__ float Bds[TILE_SIZE][TILE_SIZE];

   int tx = threadIdx.x;
   int ty = threadIdx.y;
   
   int i = blockIdx.y * TILE_SIZE + ty;
   int k = blockIdx.x * TILE_SIZE + tx;
   
   float Celem = 0;
   
   for(int m=0; m < n/TILE_SIZE; m++) {
      Ads[ty][tx] = Ad[ i*n + m*TILE_SIZE+tx];
      Bds[ty][tx] = Bd[ (m*TILE_SIZE+ty)*n + k];
      __syncthreads();
      
      for(int j=0; j<TILE_SIZE; j++)
	 Celem += Ads[ty][j]*Bds[j][tx];
   
      __syncthreads();
   };
   Cd[i*n+k] += Celem;
}


/* 
 * matrix multiplication C += A*B 
 *  -> CUDA kernel
 *     (implementation adopted from Kirk&Hwu: 
 *      "Programming Massively Parallel Processors, chapter 5)
 *  -> Features:
 *     - tiled matrix multiplication with use of shared memory
 *     - coalesced memory access
 *     - overlapping loads of subsequent tile pairs (using registers & shared memory)
 */
__global__ void matrixMultKernel_overlap(float* Ad, float* Bd, float* Cd, int n)
{
   __shared__ float Ads[TILE_SIZE][TILE_SIZE];
   __shared__ float Bds[TILE_SIZE][TILE_SIZE];

   int tx = threadIdx.x;
   int ty = threadIdx.y;
   
   int i = blockIdx.y * TILE_SIZE + ty;
   int k = blockIdx.x * TILE_SIZE + tx;
   
   float Celem = 0;
   
   float Areg = Ad[ i*n + tx];
   float Breg = Bd[ ty*n + k];
   
   for(int m=1; m < n/TILE_SIZE; m++) {
      Ads[ty][tx] = Areg;
      Bds[ty][tx] = Breg;
      __syncthreads();
      
      Areg = Ad[ i*n + m*TILE_SIZE+tx];
      Breg = Bd[ (m*TILE_SIZE+ty)*n + k];

      for(int j=0; j<TILE_SIZE; j++)
	 Celem += Ads[ty][j]*Bds[j][tx];
   
      __syncthreads();
   };
   
   Ads[ty][tx] = Areg;
   Bds[ty][tx] = Breg;
   __syncthreads();
   for(int j=0; j<TILE_SIZE; j++)
      Celem += Ads[ty][j]*Bds[j][tx];
   __syncthreads();

   Cd[i*n+k] += Celem;
}


